var path = require('path');
var json2csv = require('json2csv');
//var exportExcel = require("../components/helpers/exportODS");
var mssql = require('mssql');
var winston = require('winston');
var logger = winston.loggers.get('default');
var transactionLogger = winston.loggers.get('transaction');
var moment = require('moment');
//var sequelize = require('sequelize');
var IOService = require("../components/services/io");

var fs = require('fs');

var _ = require("lodash");

var env = process.env.NODE_ENV || "prod";
var configJson = require(path.join(__dirname, '..', '/components/config/config.json'))[env];

var config = {
    server: configJson.host,
    user: configJson.username,
    password: configJson.password,
    database: configJson.database,
    stream: true,
    port: configJson.port,
    requestTimeout: configJson.dialectOptions.requestTimeout,
    options: {
        encrypt: true
    }
};

mssql.on('error', function(err) {
    logger.error("sql error: " + err);
});

module.exports = function(app, passport, models) {

    //APIs for Mobile Application
    app.post('/api/saveCountData', function(req, res) {
        bulkCreateData(req.body, 'count')
            .done(function(data) {
                res.send(data);
            });
    });

    app.post('/api/saveInventoryData', function(req, res) {
        bulkCreateData(req.body, 'inventory')
            .done(function(data) {
                res.send(data);
            });
    });

    //apis for downloading csv files
    function filterRow(row) {
        if (row.expiration != null) {
            var date = "\"" + moment(row.expiration).format("MM/YY") + "\"";
            row.expiration = date;
        }
        if (row.cnt_ind != null) {
            if (row.cnt_ind == 1) {
                row.cnt_ind = "X";
            } else {
                row.cnt_ind = "";
            }
        }

        //strip non-numerics from clinic ID (deprecated)
        //if (row.clinicId != null) {
        //    var clinicId = row.clinicId.replace(/[^0-9]/g, '');
        //    row.clinicId = clinicId;
        //}
    }

    function filterClinicRow(row, vendors) {
        for (var field in row) {
            if (String(field) == "catalogs") {
                //strip the default vendors from the field
                for (var v=0; v < vendors.length; v++) {
                    if (String(row[field]).indexOf(",") >= 0) {
                        row[field] = String(row[field]).replace("," + vendors[v].vendorId, '');
                        row[field] = String(row[field]).replace(vendors[v].vendorId + ",", '');
                    } else {
                        row[field] = String(row[field]).replace(vendors[v].vendorId, '');
                    }
                }
            } else {
                if (String(row[field])==="false") row[field] = 'n';
                if (String(row[field])==="true") row[field] = 'y';
            }

        }
    }

    function filterClinicRowSequelize(row, vendors, fields) {
        var data = {};
        _.forEach( fields, function(field) {
            if (field == "catalogs" && row.Catalogs) {
                data[field] = _.map(row.Catalogs, 'vendorId').join(",");
                //strip the default vendors from the field
                for (var v=0; v < vendors.length; v++) {
                    if (String(data[field]).indexOf(",") >= 0) {
                        data[field] = String(data[field]).replace("," + vendors[v].vendorId, '');
                        data[field] = String(data[field]).replace(vendors[v].vendorId + ",", '');
                    } else {
                        data[field] = String(data[field]).replace(vendors[v].vendorId, '');
                    }
                }
            } else {
                data[field] = row[field];
                if (String(row[field])==="false") data[field] = 'n';
                if (String(row[field])==="true") data[field] = 'y';
            }
        });

        return data;
    }

    //Data cleanup for the Master Item Export File
    function filterItemRow(row, fields) {
        var data = {};
        _.forEach( fields, function(field) {
           data[field] = (row[field]) ? row[field] : "";
        });

        if (row.cntInd != null) {
            if (data.cntInd == 1) {
                data.cntInd = "X";
            } else {
                data.cntInd = "";
            }
        }
        data['onhand_inventory_ind'] = data.showInventoryInd;

        if (row.Catalogs) {
            data['medVendor'] = row.catalogAsString;
        }
        else {
            data['medVendor'] = "";
        }

        if (data.discontinued === null || !data.discontinued) {
            data.discontinued = "";
        } else if (row.discontinued && row.discontinued.__proto__ && row.discontinued.__proto__.isValid ) {
            if (!row.discontinued.isValid()) {
                data.discontinued = "";
            }
        }

        if (data.alternates == null) {
            data.alternates = "";
        }

        return data;
    }

    app.post('/api/getCountFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;
            request.query('select time_stamp, clinic_id, sap_material, vendor_material, material_description, count, count_type, cat_type from tbl_count_data');

            var data = [];
            request.on('row', function(row) {
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("error: " + err);
            });
            request.on('done', function(returnVal) {
                var fields = ['time_stamp', 'clinic_id', 'sap_material', 'vendor_material', 'material_description', 'count', 'count_type', 'cat_type'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);
                    res.set({
                        'Content-Disposition': 'attachment; filename=countData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getCountFile', function(req, res){
         models.CountData.findAll({ raw:true }).then(function(data){
             var fields = ['timestamp', 'clinicId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'countType', 'catType'];
             json2csv({ data: data, fields: fields }, function(err, csv) {
                 if (err) logger.error(err);
                 res.set({
                     'Content-Disposition': 'attachment; filename=countData.csv',
                     'Content-Type': 'text/csv'
                 });
                 res.end(csv);
             });
         }).catch(function(err){
             logger.error("error: " + err);
         });
    });

    app.post('/api/getInventoryFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select time_stamp, clinic_id, user_id, sap_material, vendor_material, description, count, " +
                "count_type, lot, expiration from tbl_inventory where type = 'INV'");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['time_stamp', 'clinic_id', 'user_id', 'sap_material', 'vendor_material', 'description', 'count', 'count_type', 'lot', 'expiration'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=inventoryData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getInventoryFile', function(req, res){
        models.Inventory.findAll({ where:{ type: 'INV' }, raw:true }).then(function(data){

            _.forEach( data, function(row) {
                filterRow(row);
            });

            var fields = ['timestamp', 'clinicId', 'userId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'countType', 'lot', 'expiration'];
            json2csv({ data: data, fields: fields }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=inventoryData.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/v2/getMasterItemFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            var q = "select id, sap_material_id, material_description, cnt_ind, cnt_description, " +
            " ISNULL(STUFF((SELECT ',' + v.vendor_id FROM db_owner.tbl_vendor_list v, db_owner.tbl_item_catalog ic WHERE ic.item_id = mi.id and ic.vendor_id = v.id for xml path('')),1,1,''),'') as catalogs, " +
            "    vendor_material, conv, count_type, " +
            "    ISNULL(alternates,'') as alternates, onhand_inventory_ind, " +
            "    CASE WHEN discontinued IS NULL THEN '' ELSE CONVERT(VARCHAR,discontinued,21) END as discontinued" +
            " from db_owner.tbl_master_item mi";
            logger.debug(q);
            request.query(q);
            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['sap_material_id', 'material_description', 'cnt_ind', 'cnt_description', 'catalogs', 'vendor_material', 'conv', 'count_type', 'alternates', 'onhand_inventory_ind', 'discontinued'];
                json2csv({ data: data, fields: fields, del: '\t' }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=inventoryData.csv',
                        'Content-Type': 'text/tab-separated-values'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v3/getMasterItemFile', function(req, res){
        models.MasterItem.findAll({ include: [{ model: models.VendorList, as: 'Catalogs' }],
            paranoid:false }).then(function(data){
            var cleanData = [];
            var fields = ['sapMaterial', 'description', 'cntInd', 'cntDescription', 'medVendor', 'vendorMaterial', 'conv', 'countType', 'alternates', 'onhand_inventory_ind', 'discontinued'];

            _.forEach( data, function(row) {
                cleanData.push(filterItemRow(row, fields));
            });

            json2csv({ data: cleanData, fields: fields, del: '\t' }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=itemData.tsv',
                    'Content-Type': 'text/tsv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/getMasterItemFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select sap_material_id, material_description, cnt_ind, cnt_description, med_vendor, vendor_material, conv, " +
                "count_type, CASE WHEN alternates IS NULL THEN '' ELSE alternates END as alternates, onhand_inventory_ind, CASE WHEN discontinued IS NULL THEN '' ELSE CONVERT(VARCHAR,discontinued,21) END as discontinued from tbl_master_item mi");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['sap_material_id', 'material_description', 'cnt_ind', 'cnt_description', 'med_vendor', 'vendor_material', 'conv', 'count_type', 'alternates', 'onhand_inventory_ind', 'discontinued'];
                json2csv({ data: data, fields: fields, del: '\t' }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=inventoryData.csv',
                        'Content-Type': 'text/tab-separated-values'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getClinicProfileFile', function(req, res) {
        models.VendorList.findAll({ where: {isDefault: true} }).then(function(vendors) {
            var conn = new mssql.Connection(config, function(err) {
                conn.connect();
                var request = new mssql.Request(conn);
                request.stream = true;
                var q = "SELECT clinic_id, ISNULL(STUFF((SELECT ',' + v.vendor_id FROM db_owner.tbl_vendor_list v, db_owner.tbl_clinic_catalog cc WHERE cc.clinic_id = cp.id and cc.vendor_id = v.id for xml path('')),1,1,''),'') as catalogs, " +
                    " mircera_visible, mircera_io, areanesp_visible as aranesp_visible, areanesp_io as aranesp_io, epogen_visible, epogen_io, count_in_each " +
                    " from tbl_clinic_profile cp order by clinic_id asc";
                logger.debug(q);
                request.query(q);

                var data = [];
                var idx = 0;
                var defVendorField = 'default_catalogs';
                var instructionsField = "For best results, open a blank worksheet and import this file";
                request.on('row', function(row) {

                    filterClinicRow(row, vendors);
                    if (idx < vendors.length) {
                        row[defVendorField] = vendors[idx].vendorId;
                    } else {
                        row[defVendorField] = "";
                    }
                    idx++;
                    data.push(row);
                });
                request.on('error', function(err) {
                    logger.error("error: " + err);
                });
                request.on('done', function(returnVal) {

                    var fields = ['clinic_id', 'catalogs', 'mircera_visible', 'mircera_io', 'aranesp_visible', 'aranesp_io', 'epogen_visible', 'epogen_io', 'count_in_each', defVendorField, '', instructionsField];
                    json2csv({ data: data, fields: fields, del: '\t' }, function(err, csv) {
                        if (err) logger.error(err);

                        res.set({
                            'Content-Disposition': 'attachment; filename=CHIPS_Locator_Data.tsv',
                            'Content-Type': 'text/tab-separated-values'
                        });
                        res.end(csv);
                    });
                });
            });
        });
    });

    app.post('/api/v3/getClinicProfileFile', function(req, res) {
        models.VendorList.findAll({ where: {isDefault: true} }).then(function(vendors) {
            models.ClinicProfile.findAll({
                include: [{ model: models.VendorList, as: 'Catalogs' }],
                order: [
                    ['clinicId', 'ASC']]}).then(function(clinics) {

                var defVendorField = 'default_catalogs';
                var instructionsField = "For best results, open a blank worksheet and import this file";
                var cleanData = [];
                var fields = ['clinicId', 'catalogs', 'mirceraVisible', 'mirceraIo', 'areanespVisible', 'areanespIo', 'epogenVisible', 'epogenIo', 'countInEach'];
                for (var idx = 0; idx < clinics.length; idx++) {
                    var row = clinics[idx];
                    var newRow = filterClinicRowSequelize(row, vendors, fields);
                    if (idx < vendors.length) {
                        newRow[defVendorField] = vendors[idx].vendorId;
                    } else {
                        newRow[defVendorField] = "";
                    }
                    cleanData.push(newRow);
                }
                fields = _.concat(fields, defVendorField, '', instructionsField);
                json2csv({ data: cleanData, fields: fields, del: '\t' }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=CHIPS_Locator_Data.tsv',
                        'Content-Type': 'text/tab-separated-values'
                    });
                    res.end(csv);
                });

            }).catch(function(cerr){
                logger.error(cerr);
            });
        }).catch(function(verr){
            logger.error(verr);
        });
    });

    app.post('/api/getClinicProfileFile', function(req, res) {
        models.VendorList.findAll({ where: {isDefault: true} }).then(function(vendors) {
            var conn = new mssql.Connection(config, function(err) {
                conn.connect();
                var request = new mssql.Request(conn);
                request.stream = true;

                request.query("SELECT clinic_id, vendor_list, mircera_visible, mircera_io, areanesp_visible as aranesp_visible, areanesp_io as aranesp_io, epogen_visible, epogen_io, count_in_each " +
                    " from tbl_clinic_profile  order by clinic_id asc");

                var data = [];
                var idx = 0;
                var defVendorField = 'default_vendors';
                var instructionsField = "For best results, open a blank worksheet and import this file";
                request.on('row', function(row) {

                    filterClinicRow(row, vendors);
                    if (idx < vendors.length) {
                        row[defVendorField] = vendors[idx].vendorId;
                    } else {
                        row[defVendorField] = "";
                    }
                    idx++;
                    data.push(row);
                });
                request.on('error', function(err) {
                    logger.error("error: " + err);
                });
                request.on('done', function(returnVal) {

                    var fields = ['clinic_id', 'vendor_list', 'mircera_visible', 'mircera_io', 'aranesp_visible', 'aranesp_io', 'epogen_visible', 'epogen_io', 'count_in_each', defVendorField, '', instructionsField];
                    json2csv({ data: data, fields: fields, del: '\t' }, function(err, csv) {
                        if (err) logger.error(err);

                        res.set({
                            'Content-Disposition': 'attachment; filename=CHIPS_Locator_Data.tsv',
                            'Content-Type': 'text/tab-separated-values'
                        });
                        res.end(csv);
                    });
                });
            });
        });
    });

    app.post('/api/getPoFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select time_stamp, clinic_id, user_id, sap_material, vendor_material, description, count, " +
                "count_type, lot, expiration from tbl_inventory where type = 'PO'");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("request error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['time_stamp', 'clinic_id', 'user_id', 'sap_material', 'vendor_material', 'description', 'count', 'count_type', 'lot', 'expiration'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=inventoryData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getPoFile', function(req, res){
        models.Inventory.findAll({ where:{ type: 'PO' }, raw:true }).then(function(data){

            _.forEach( data, function(row) {
                filterRow(row);
            });

            var fields = ['timestamp', 'clinicId', 'userId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'countType', 'lot', 'expiration'];
            json2csv({ data: data, fields: fields }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=poData.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/getReceiptFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select time_stamp, clinic_id, user_id, sap_material, vendor_material, description, count, " +
                "po_prefix, po_number, lot, expiration from tbl_inventory where type = 'RECEIPT'");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("request error: " + err);
            });
            request.on('done', function(returnVal) {
                var fields = ['time_stamp', 'clinic_id', 'user_id', 'sap_material', 'vendor_material', 'description', 'count', 'po_prefix', 'po_number', 'lot', 'expiration'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=receiptData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getReceiptFile', function(req, res){
        models.Inventory.findAll({ where:{ type: 'RECEIPT' }, raw:true }).then(function(data){

            _.forEach( data, function(row) {
                filterRow(row);
            });

            var fields = ['timestamp', 'clinicId', 'userId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'poPrefix', 'poNumber', 'lot', 'expiration'];
            json2csv({ data: data, fields: fields }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=receiptData.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });

    app.post('/api/getConsumptionFile', function(req, res) {
        var conn = new mssql.Connection(config, function(err) {
            conn.connect();
            var request = new mssql.Request(conn);
            request.stream = true;

            request.query("select time_stamp, clinic_id, user_id, sap_material, vendor_material, description, count, " +
                "lot, expiration from tbl_inventory where type = 'CONSUMERUSAGE'");

            var data = [];
            request.on('row', function(row) {
                filterRow(row);
                data.push(row);
            });
            request.on('error', function(err) {
                logger.error("request error: " + err);
            });
            request.on('done', function(returnVal) {

                var fields = ['time_stamp', 'clinic_id', 'user_id', 'sap_material', 'vendor_material', 'description', 'count', 'lot', 'expiration'];
                json2csv({ data: data, fields: fields }, function(err, csv) {
                    if (err) logger.error(err);

                    res.set({
                        'Content-Disposition': 'attachment; filename=consumptionData.csv',
                        'Content-Type': 'text/csv'
                    });
                    res.end(csv);
                });
            });
        });
    });

    app.post('/api/v2/getConsumptionFile', function(req, res){
        models.Inventory.findAll({ where:{ type: 'CONSUMERUSAGE' }, raw:true }).then(function(data){

            _.forEach( data, function(row) {
                filterRow(row);
            });

            var fields = ['timestamp', 'clinicId', 'userId', 'sapMaterial', 'vendorMaterial', 'description', 'count', 'countType', 'lot', 'expiration'];
            json2csv({ data: data, fields: fields }, function(err, csv) {
                if (err) logger.error(err);
                res.set({
                    'Content-Disposition': 'attachment; filename=consumptionData.csv',
                    'Content-Type': 'text/csv'
                });
                res.end(csv);
            });
        }).catch(function(err){
            logger.error("error: " + err);
        });
    });


    //MASTER ITEMs
    //...this is used by devices
    app.get('/api/getMasterItemList', function(req, res) {
        commonMasterItemList(req, res, true);
    });

    app.get('/api/v2/getMasterItemList', function(req, res) {
        commonMasterItemList(req, res, false);
    });

    function commonMasterItemList(req, res, isBackwards) {
        var date = req.query.date;
        var where = [];
        //logger.debug("getting master items at: " + new Date());

        if (date != null) {
            //logger.debug("getting master items for date: " + JSON.stringify(date));
            where.push({date_modified: {$gt: date}});
        }
        models.MasterItem.findAll({
                include: [{ model: models.VendorList, as: 'Catalogs' }],
                where: where,
                paranoid: false
            }
        ).then(function (countRecords) {
            updateItemCatalog(countRecords, isBackwards);
            //logger.debug("sending master items: " + new Date());
            logger.log('trace', JSON.stringify(countRecords)) ;
            res.send(countRecords);
        });
    }

    //...this is used by the UI
    app.post('/api/getMasterItem', function(req, res) {
        if (req.body.sapMaterial == null) {
            models.MasterItem.findAll({ include: [{ model: models.VendorList, as: 'Catalogs' }] })
                .then(function (items) {
                    updateItemCatalog(items);
                    res.send(items);

                });
        } else {
            models.MasterItem.findAll({
                include: [{ model: models.VendorList, as: 'Catalogs' }],
                where: {
                    $and: [
                        {sapMaterial: req.body.sapMaterial},
                        {vendorMaterial: req.body.vendorMaterial}
                    ]
                }
            }).then(function (items) {
                updateItemCatalog(items, true);
                res.send(items);
            });
        }
    });

    //...this is used by the UI
    app.post('/api/v2/getMasterItem', function(req, res) {
        if (req.body.sapMaterial == null) {
            if (req.body.where) {
                var args = buildMasterItemWhereArgs(req);
                models.MasterItem.findAll({ where: args, include: [{model: models.VendorList, as: 'Catalogs'}]})
                    .then(function (items) {
                        updateItemCatalog(items);
                        res.send(items);

                    });
            } else {
                models.MasterItem.findAll({include: [{model: models.VendorList, as: 'Catalogs'}]})
                    .then(function (items) {
                        updateItemCatalog(items);
                        res.send(items);

                    });
            }
        } else {
            models.MasterItem.findAll({
                include: [{ model: models.VendorList, as: 'Catalogs' }],
                where: {
                    $and: [
                        {sapMaterial: req.body.sapMaterial},
                        {vendorMaterial: req.body.vendorMaterial}
                    ]
                }
            }).then(function (items) {
                updateItemCatalog(items, false);
                res.send(items);
            });
        }
    });

    function updateItemCatalog(items, isBackwards) {
        for (var i=0; i < items.length; i++) {
            var item = items[i];
            item.medVendor = "";
            var catalogs = item.Catalogs;

            if (isBackwards) {
                if (catalogs && catalogs.length > 0) {
                    item.medVendor = catalogs[0].vendorId;
                }
            } else {
                item.medVendor = item.catalogAsString;
            }
            //delete item.Catalogs
        }
        return items;
    }

    //CLINICS
    //this is used by devices.... but deprecated
    app.get('/api/getMobileClinicProfile', function(req, res) {
        var clinicId = req.query.clinic;
        var date = req.query.date;

        var where = [];
        if (date != null) {
            //logger.debug("getting clinic profile for date modified: " + JSON.stringify(date));
            where.push({date_modified: {$gt: date}});
        }
        if (clinicId != null) {
            // where.push(  );

            models.ClinicProfile.findAll({
                where: models.ClinicProfile.sequelize.where(models.ClinicProfile.sequelize.fn('dbo.fnRemoveNonNumericCharacters', models.ClinicProfile.sequelize.col('clinic_id')), clinicId)
                ,raw : true
            }).then(function (profile) {
                //old Android code is expecting one row with a numeric clinic id
                //var json = JSON.parse(JSON.stringify(profile));
                if (profile[0] != null) {
                    profile[0].clinicId = new Number(profile[0].clinicId.replace(/[^0-9]+/g, ''));
                    res.send(profile[0]);
                } else {
                    res.sendStatus(500);
                }
            });
        } else {
            res.sendStatus(500);
        }
    });

    //this is used by devices... and must be backwards compatible
    app.get('/api/v2/getMobileClinicProfile', function(req, res) {
        var clinicId = req.query.clinic;
        var date = req.query.date;

        var where = [];
        if (date != null) {
            //logger.debug("getting clinic profile for date modified: " + JSON.stringify(date));
            where.push({date_modified: {$gt: date}});
        }
        if (clinicId != null) {
            where.push({clinic_id: clinicId});
            models.ClinicProfile.findOne({
                include: [{ model: models.VendorList, as: 'Catalogs' }],
                where: where
            }).then(function (profile) {
                updateClinicVendorList(profile);
                res.send(profile)
            });
        } else {
            res.sendStatus(500);
        }
    });


    function updateClinicCatalog(clinics) {
        for (var i=0; i < clinics.length; i++) {
            var clinic = clinics[i];
            updateClinicVendorList(clinic);
        }
        return clinics;
    }

    function updateClinicVendorList(clinic){
        clinic.vendorList = "";
        var catalogs = clinic.Catalogs;
        var v = [];
        for (var c = 0; c < catalogs.length; c++){
            v.push(catalogs[c].vendorId);
        }
        clinic.vendorList = v.join();
    }

    function buildDeviceCountQuery(location) {
        var table_name = "[tbl_master_item]";
        var vendor_fld = "[vendor_id]";
        var desc_fld="[material_description]";
        var cntType_fld = "[count_type]";
        var vl = location.vendorList;
        var vlList = [];
        if (vl && vl.length > 0 && vl.indexOf(",") > -1) {
            vlList = vl.split(/,/);
        }
        var where = [];
        //first vendors
        if (vlList.length == 0) {
            where.push(table_name + "."+ vendor_fld + " = ''");
        }
        for (var v = 0; v < vlList.length; v++ ) {
            var vendor = vlList[v];
            if (!vendor || vendor.length == 0) {
                where.push("vl."+ vendor_fld + " = ''")
            } else {
                var escapedName = vendor; //models.sequelize.constructor.Utils.escape(vendor);
                where.push("vl." + vendor_fld + " = '" + escapedName + "'");
            }
        };
        var query = 'SELECT * FROM ' + table_name + " mi" +
            " WHERE mi.discontinued is null AND mi.[cnt_ind] = 1 AND mi." + cntType_fld + " IN ('ESA','Other') AND EXISTS (SELECT 1 FROM tbl_item_catalog ic, tbl_vendor_list vl WHERE ic.item_id = mi.id AND vl.id = ic.vendor_id AND (";
        for (var v=0; v < where.length; v++) {
            var w = where[v];
            query += ((v==0) ? "" : " OR ") + w;
        };
        query += "))";
        where=[];
        //now item names
        if (!(location.mirceraVisible===true)) {
            where.push("mi." + desc_fld + " NOT LIKE 'MIRCERA%' ");
        }
        if (!(location.epogenVisible===true)) {
            where.push("mi." + desc_fld + " NOT LIKE 'EPOGEN%' ");
        }
        if (!(location.areanespVisible===true)) {
            where.push("mi." + desc_fld + " NOT LIKE 'ARANESP%' ");
        }
        if (where.length > 0) {
            query += " AND ( " + where[0];
            if (where.length > 1) {
                for (var w = 1; w < where.length; w++) {
                    query += " AND " + where[w];
                }
            }
            query += ")";
        }
        query += " ORDER BY mi." + cntType_fld + ", mi." + desc_fld;
        console.log(query);
        return query;
    }

    app.post('/api/getMasterItemAsDeviceCount', function(req, res) {
        var location = req.body.location;
        if (location) {
            var query = buildDeviceCountQuery(location);
            models.sequelize.query(query, { model: models.MasterItem}).then(function (items) {
                res.send(items);
            });
        }

    });

    app.post('/api/v2/getMasterItemAsDeviceCount', function(req, res) {
        var location = req.body.location;
        if (location) {
            var query = buildDeviceCountQuery(location);
            models.sequelize.query(query, { model: models.MasterItem}).then(function (items) {
                var io = new IOService(configJson);
                var promises = [];
                var clinic = location.clinicId;
                for (var i=0; i< items.length; i++) {
                    var item = items[i];
                    var vendorMaterial = item.dataValues.vendor_material;
                    promises.push(io.getParValuePromise(clinic, vendorMaterial));
                    item.dataValues["min_val"] = 0;
                    item.dataValues["max_val"] = 0;
                }  
                Promise.all(promises)
                    .catch(function(err){
                        
                    }).then(function(results){
                        for(var r = 0; r< results.length; r++) {
                            var par = io.processResponse(results[r]);
                            if (par) {
                                var item = null;
                                if (par.vendorMaterial) {
                                    for (var i = 0; i < items.length; i++) {
                                        if (items[i].dataValues.vendor_material == par.vendorMaterial) {
                                            item = items[i];
                                            break;
                                        }
                                    }
                                    if (item) {
                                        item.dataValues["min_val"] = par.minVal;
                                        item.dataValues["max_val"] = par.maxVal;
                                    }
                                }
                            }
                        }
                        res.send(items);
                })
            });
        }

    });



    //Special API for add or update functionality
    app.post('/api/v2/saveMasterItem', function(req, res) {
        console.log("request: " + req.session);
        upsertMasterItem(req.body, req)
            .then(function(result) {
                if (!result) result = req.body;
                res.send(result);
            });
    });

    app.post('/api/deleteMasterItem', function(req, res) {
        models.MasterItem.destroy({
            where: { id: req.body.id }
        })
            .then(function (){
                transactionLogger.info("Master item deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
                res.send(req.body);
            });
    });

    app.post('/api/getClinicProfile', function(req, res) {
        if (req.body.clinicId == null) {
            models.ClinicProfile.findAll().then(function(clinicProfiles) {
                res.send(clinicProfiles);
            });
        } else {
            models.ClinicProfile.findAll({
                where: {
                    clinicId: req.body.clinicId
                }

            }).then(function (clinicProfiles) {
                res.send(clinicProfiles);
            });
        }
    });

    app.post('/api/v2/getClinicProfile', function(req, res) {
        if (req.body.clinicId == null) {
            if (req.body.where) {
                var args = buildClinicProfileWhereArgs(req);
                models.ClinicProfile.findAll({ where: args, include: [{model: models.VendorList, as: 'Catalogs'}]})
                    .then(function (clinicProfiles) {
                        updateClinicCatalog(clinicProfiles);
                        res.send(clinicProfiles);
                    });
            } else {
                models.ClinicProfile.findAll({
                    include: [{
                        model: models.VendorList,
                        as: 'Catalogs'
                    }]
                }).then(function (clinicProfiles) {
                    updateClinicCatalog(clinicProfiles);
                    res.send(clinicProfiles);
                });
            }
        } else {
            models.ClinicProfile.findAll({
                include: [{ model: models.VendorList, as: 'Catalogs' }],
                where: {
                    clinicId: req.body.clinicId
                }

            }).then(function (clinicProfiles) {
                updateClinicCatalog(clinicProfiles);
                res.send(clinicProfiles);
            });
        }
    });

    app.post('/api/deleteClinicProfile', function(req, res) {
        models.ClinicProfile.destroy(
            {where: { id: req.body.id }}
        ).then(function () {
                transactionLogger.info("Clinic profile deleted by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(req.body));
                res.send(req.body);
            });
    });

    app.post('/api/v2/saveClinicProfile', function(req, res) {
        upsertClinicProfile(req.body, req)
            .then(function(result) {
                if (!result) result = req.body;
                res.send(result);
            });
    });

    //IS THIS EVEN USED????
    app.post('/api/saveLogData', function(req, res) {
        var msg = "";
        try {
            msg = JSON.stringify(req.body);
        } catch (e) {
            msg = "stringify error";
        }
        winston.debug("log from api: " + msg);
        res.sendStatus(200);
    });

    /**
     * Adds a specified vendorId to every clinic's list
     *   that does not already have it.
     *
     * NOTE: The Vendor ID MUST exist in the Vendor table
     *
     * Sample request bodies: { "vendorId": "value" }   OR   { "vendorId": "-BLANK-" }
    **/
    app.post('/api/bulkAddClinicVendor', function(req, res) {
        console.log("bulkAddClinicVendor  BODY:" + req.body.vendorId);
        models.VendorList.findAll({
            where: {
                $or: [
                    { vendorId : req.body.vendorId},
                    { vendorDisplay: req.body.vendorId }
                ]
            },raw:true})
            .then(function(vendors) {
                if (vendors && vendors.length > 0) {
                    var table_name = "[tbl_clinic_profile]";
                    var vendor_fld = "[vendor_list]";
                    var query = "UPDATE " + table_name + " SET " + vendor_fld + " = " + vendor_fld + " + '," + vendors[0].vendorId + "'";
                    query += " WHERE " + vendor_fld + " NOT LIKE '" + vendors[0].vendorId + ",%'"; //starts with
                    query += " AND " + vendor_fld + " NOT LIKE '%," + vendors[0].vendorId + "'"; //ends with
                    query += " AND " + vendor_fld + " NOT LIKE '%," + vendors[0].vendorId + ",%'"; //contains middle
                    query += " AND " + vendor_fld + " <> '" + vendors[0].vendorId + "'"; //only entry
                    console.log(query);
                    models.sequelize.query(query).spread(function(results, metadata) {
                        res.send(results);
                    });
                } else {
                    res.sendStatus(500);
                }
            }, function(err) {
                logger.error(err);
                res.status = 500;
                res.send(err);
            });
    });

    app.post('/api/v2/bulkAddClinicVendor', function(req, res) {
        console.log("bulkAddClinicVendor  BODY:" + req.body.vendorId);
        models.VendorList.findAll({
            where: {
                $or: [
                    { vendorId : req.body.vendorId},
                    { vendorDisplay: req.body.vendorId }
                ]
            },raw:true})
            .then(function(vendors) {
                if (vendors && vendors.length > 0) {
                    var table_name = "[tbl_clinic_catalog]";
                    var clinic_tbl = "[tbl_clinic_profile]";
                    for (var v=0; v < vendors.length; v++) {
                        var vendor = vendors[v];
                        var reset = "DELETE FROM " + table_name + " WHERE vendor_id = " + vendor.id;
                        var query = "INSERT INTO " + table_name + " (clinic_id, vendor_id) SELECT ID, " + vendor.id + " FROM  " + clinic_tbl;
                        console.log(query);
                        models.sequelize.query(reset).spread(function(resets, metadata) {
                            models.sequelize.query(query).spread(function(results, metadata) {
                                res.send(results);
                            }).catch(function(insErr) {
                                logger.error(insErr);
                                res.status = 500;
                                res.send(insErr);
                            });
                        }).catch(function(delErr) {
                            logger.error(delErr);
                            res.status = 500;
                            res.send(delErr);
                        });
                    }
                } else {
                    res.sendStatus(500);
                }
            }).catch(function(err) {
                logger.error(err);
                res.status = 500;
                res.send(err);
            });
    });

    /**
     * Removes a specified vendorId in every clinic's list
     *   that already has it
     *
     * NOTE: The Vendor ID is NOT checked against the Vendor table
     *
     * Sample request bodies: { "vendorId": "value" }   OR   { "vendorId": "-BLANK-" }
     **/
    app.post('/api/bulkRemoveClinicVendor', function(req, res) {
        if (req.body && req.body.vendorId) {
            if (req.body.vendorId) {
                var vendor = req.body.vendorId;
                if (vendor == "-BLANK-") { vendor = ""; }
                var table_name = "[tbl_clinic_profile]";
                var vendor_fld = "[vendor_list]";
                var query = "UPDATE " + table_name + " SET " + vendor_fld + " = (CASE  ";
                query += " WHEN " + vendor_fld + " LIKE '" + vendor + ",%'   THEN REPLACE(" + vendor_fld + ", '" + vendor + ",','') "; //starts with
                query += " WHEN " + vendor_fld + " LIKE '%," + vendor + "'   THEN REPLACE(" + vendor_fld + ", '," + vendor + "','') "; //ends with
                query += " WHEN " + vendor_fld + " LIKE '%," + vendor + ",%' THEN REPLACE(" + vendor_fld + ", '," + vendor + "','') "; //contains middle
                query += " WHEN " + vendor_fld + " = '" + vendor + "' THEN '' "; //only entry
                query += " ELSE " + vendor_fld + " END) ";
                query += " WHERE " + vendor_fld + " LIKE '" + vendor + ",%'"; //starts with
                query += " OR " + vendor_fld + "  LIKE '%," + vendor + "'"; //ends with
                query += " OR " + vendor_fld + "  LIKE '%," + vendor + ",%'"; //contains middle
                query += " OR " + vendor_fld + " = '" + vendor + "'"; //only entry

                models.sequelize.query(query).spread(function(results, metadata) {
                    res.send(results);
                }, function(err) {
                    logger.error(err);
                    res.status = 500;
                    res.send(err);
                });
            } else {
                res.sendStatus(500);
            }
        };
    });

    /**
     * Removes a specified vendorId in every clinic's list
     *   that already has it
     *
     * NOTE: The Vendor ID MUST BE checked against the Vendor table
     *
     * Sample request bodies: { "vendorId": "value" }   OR   { "vendorId": "-BLANK-" }
     **/
    app.post('/api/v2/bulkRemoveClinicVendor', function(req, res) {
        if (req.body && req.body.vendorId) {
            if (req.body.vendorId) {
                var vendor = req.body.vendorId;
                var table_name = "[tbl_clinic_catalog]";
                var vendor_table_name = "[tbl_vendor_list]";
                var vendor_fld = "[vendor_id]";
                var query = "DELETE FROM " + table_name;
                query += " WHERE " + vendor_fld + " IN (SELECT ID FROM " + vendor_table_name + " WHERE " + vendor_fld + " = '" + vendor + "'";
                query += " OR [vendor_display] = '" + vendor + "')";

                models.sequelize.query(query).spread(function(results, metadata) {
                    res.send(results);
                }).catch(function(err) {
                    logger.error(err);
                    res.status = 500;
                    res.send(err);
                });
            } else {
                res.sendStatus(500);
            }
        };
    });

    /**
     *  Update or Insert a record. Searches for a record, if found, updates it. If not found, creates it then updates it.
     *  Returns 1 if updated, 0 if created, -1 for error (but not really)
     */
    function upsertClinicProfile(clinicProfile, req) {
        if (clinicProfile.hasOwnProperty("Catalogs")) delete clinicProfile.Catalogs;
        return models.ClinicProfile.findOrCreate({where: {clinicId: clinicProfile.clinicId}, raw: true})
            .spread(function(clinicRecord, created) {
                clinicProfile.id = clinicRecord.id;
                models.ClinicProfile.update(clinicProfile, { where: { id: clinicProfile.id }})
                    .then(function(count) {
                            if (count > 0 && clinicProfile.vendorList) {
                                var ids = clinicProfile.vendorList.split(",");
                                models.ClinicCatalog.destroy({ where: { clinicId: clinicProfile.id }}).then(function(num){
                                    var cats = [];
                                    for (var i = 0; i < ids.length;i++) {
                                        var cat = models.ClinicCatalog.build({
                                            clinicId: clinicProfile.id,
                                            vendorId: ids[i]
                                        });
                                        cats.push(cat);
                                    }
                                    if (cats.length > 0) {
                                        var promises = [];

                                        cats.forEach(function(cat) {
                                            promises.push(cat.save());
                                        });

                                        models.sequelize.Promise.all(promises).then(function(vals){
                                            transactionLogger.info("Clinic profile " + ((created) ? "created" : "updated") + " by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(clinicProfile));
                                            return cats.length;
                                        }).catch(function(saveErr){
                                            logger.error(saveErr);
                                            return -1;
                                        });
                                    }
                                })
                            } else {
                                transactionLogger.info("Clinic profile " + ((created) ? "created" : "updated") + " by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(clinicProfile));
                                return count;
                            }
                        }).catch(function(err) {
                            logger.error(err);
                            return -1;
                    });
            }).catch(function(err) {
                logger.error(err);
                return -1;
            });
    }

    /**
     *  Update or Insert a record. Searches for a record, if found, updates it. If not found, creates it then updates it.
     *  Returns 1 if updated, 0 if created, -1 for error (but not really)
     */
    function upsertMasterItem(masterItem, req) {
        if (masterItem.hasOwnProperty("Catalogs")) delete masterItem.Catalogs;
        return models.MasterItem.findOrCreate({where: {sapMaterial: masterItem.sapMaterial, vendorMaterial: masterItem.vendorMaterial}, raw: true, paranoid: false })
            .spread(function(itemRecord, created) {
                masterItem.id = itemRecord.id;
                models.MasterItem.update(masterItem, {where: {id: masterItem.id}})
                    .then(function (count) {
                        if (count > 0 && masterItem.medVendor) {
                            var ids = masterItem.medVendor.split(",");
                            models.ItemCatalog.destroy({where: {itemId: masterItem.id}}).then(function (num) {
                                var cats = [];
                                for (var i = 0; i < ids.length; i++) {
                                    var cat = models.ItemCatalog.build({
                                        itemId: masterItem.id,
                                        vendorId: ids[i]
                                    });
                                    cats.push(cat);
                                }
                                if (cats.length > 0) {
                                    var promises = [];

                                    cats.forEach(function (cat) {
                                        promises.push(cat.save());
                                    });

                                    models.sequelize.Promise.all(promises).then(function (vals) {
                                        transactionLogger.info("Master Item " + ((created) ? "created" : "updated") + " by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(masterItem));
                                        return cats.length;
                                    }).catch(function (saveErr) {
                                        logger.error(saveErr);
                                        return -1;
                                    });
                                }
                            }).catch(function (err) {
                                logger.error(err);
                                return -1;
                            });
                        }
                    }).catch(function (err) {
                        logger.error(err);
                        return -1;
                    });
            }).catch(function(err) {
                logger.error(err);
                return -1;
            });
    }

    /**
     *  Update or Insert each inventory or count record. Searches for a record, if found, updates it. If not found, creates it then updates it.
     *  Returns all promises (to ensure caller waits for it to finish)
     */
    function bulkCreateData(dataObjects, type) {

        var promises = [];

        if (type == 'inventory') {
            if (dataObjects != null) {
                dataObjects.forEach(function(record) {
                    transactionLogger.debug("saving inventory record: " + JSON.stringify(record));
                    record.id = null;
                    promises.push(models.Inventory.findOrCreate({where: {deviceId: record.deviceId, timestamp: record.timestamp}, raw: true})
                        .spread(function(inv, created) {
                            record.id = inv.id;
                            models.Inventory.update(record, { where: { id: record.id }});
                            return {
                                id: record.id,
                                devicePk: record.devicePk
                                };
                            },
                            function(err) {
                                logger.error(err);
                            }
                        )
                    );
                });
            }
        } else if (type == 'count') {
            if (dataObjects != null) {
                dataObjects.forEach(function (countData) {
                    transactionLogger.debug("saving count record: " + JSON.stringify(countData));
                    countData.id = null;
                    promises.push(models.CountData.findOrCreate({
                            where: {
                                deviceId: countData.deviceId,
                                timestamp: countData.timestamp
                            }, raw: true
                        }).spread(function (cd, created) {
                                countData.id = cd.id;
                                models.CountData.update(countData, {where: {id: countData.id}});
                                return {
                                    id: countData.id,
                                    devicePk: countData.devicePk
                                };
                            },
                            function (err) {
                                logger.error(err);
                            }
                        )
                    );
                });
            }
        }
        return models.sequelize.Promise.all(promises);
    }

    /**
     * Processes FTP log JSON as a list
     *
     * @param dataObjects
     * @returns {*}
     */
    function bulkCreateFTPData(dataObjects) {

        var promises = [];

        if (dataObjects != null) {
            dataObjects.forEach(function (ftpData) {
                //logger.debug("saving ftp record: " + JSON.stringify(ftpData));
                ftpData.id = null;
                promises.push(models.FTPLog.findOrCreate({
                        where: {
                            deviceId: ftpData.deviceId,
                            sent: ftpData.sent,
                            userId: ftpData.userId,
                            success: ftpData.success
                        }, raw: true
                    })
                        .spread(function (fd, created) {
                                ftpData.id = fd.id;
                                models.FTPLog.update(ftpData, {where: {id: ftpData.id}});
                                return {id : ftpData.id};
                            },
                            function (err) {
                                logger.error(err);
                            }
                        )
                );
            });
        }
        return models.sequelize.Promise.all(promises);
    }

    app.post('/api/saveFTPData', function(req, res) {
        bulkCreateFTPData(req.body)
            .done(function(data) {
                res.send(data);
            });
    });

    app.post('/api/saveVendor', function(req, res) {
        if (req.body.id && req.body.id > 0) {
            models.VendorList.update(req.body, {where: {
                id: req.body.id
            }}).then(function(count){
                res.send(req.body);
            }).catch(function (err) {
                logger.error(err);
                res.sendStatus(500);
            });
        } else {
            if (req.body.vendorId) {
                models.VendorList.findOrCreate({
                    where: {
                        vendorId: req.body.vendorId
                    }, defaults: {
                        vendorDisplay: req.body.vendorId,
                        isDefault: req.body.isDefault
                    }
                }).spread(function (data, created) {
                    if (!created) {
                        data.isDefault = req.body.isDefault;
                        data.vendorDisplay = req.body.vendorId;
                        data.save().then(function () {
                            res.send(data);
                        }).catch(function (error) {
                            logger.error(error);
                            res.sendStatus(500);
                        });
                    } else {
                        res.send(data);
                    }
                }).catch(function (err) {
                    logger.error(err);
                    res.sendStatus(500);
                });
            } else {
                res.sendStatus(500);
            }
        }
    });

    app.post('/api/deleteVendor', function(req, res) {
        if (req.body.vendorId) {
            //TODO: make sure no Clinic or Item has this vendor assigned
            models.VendorList.destroy({
                where: req.body
            })
                .then(function (){
                    transactionLogger.info("Vendor item deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
                    res.send(req.body);
                });
        }
        else {
            res.sendStatus(500);
        }
    });


    /**   MESSAGING   **/
    app.get('/api/messages/:clinic', function(req, res) {
        if (req.params && req.params['clinic']){
            var inClinic = req.params['clinic'];
            models.ClinicProfile.findOne({
                where: { clinicId: inClinic }
            }).then(function(profile){
                if (profile && profile.id) {
                    var now = new Date();
                    models.MsgRecipient.findAll({
                        include: [{model: models.Message, as: 'message',
                        where: {
                            $or: [
                                { keepAliveStart: { $eq: null } },
                                {
                                    $and: {
                                        keepAliveStart: {$lte: now},
                                        keepAliveEnd: {$gte: now}
                                    }
                                }]
                        }
                        }],
                        where: {
                            clinicId: profile.id
                        }
                    }).then(function (messages) {
                        var responseList = [];
                        if (messages && messages.length > 0) {
                            for (var r = 0; r < messages.length; r++) {
                                var item = messages[r];
                                if (item.timestamp || item.status === 'read') {
                                    if (!item.message.keepAliveStart) continue;
                                }
                                responseList.push(item.message);
                                item.status = "read";
                                item.timestamp = new moment();
                                item.save().catch(function(saveErr){
                                    console.log(saveErr);
                                });
                            }
                        }
                        res.send(responseList);
                    }).catch(function (msgErr) {
                        res.status(500).send("Error querying messages: " + msgErr);
                    });
                } else {
                    res.status(404).send("Clinic not found");
                }
            }).catch(function(clinicErr){
                res.status(404).send("Clinic not found");
            })
        } else {
            res.sendStatus(400);
        }
    });

    //create a message
    app.post('/api/message', function(req, res) {

        var inJson = req.body;
        var outJson = {};
        if (inJson.id) {
            res.status("400").send("Invalid method")
        }
        else if (!inJson.recipients || !Array.isArray(inJson.recipients)) {
            res.status("400").send("At least one recipient is required")
        }
        else {
            var msg = models.Message.build({
                message: inJson.message,
                timestamp: new Date(),
                createdByUserId: inJson.createdByUserId,
                keepAliveStart: inJson.keepAliveStart,
                keepAliveEnd: inJson.keepAliveEnd
            });

            msg.save().then(function(savedMsg){
                outJson = savedMsg;
                var promises = [];
                var id = savedMsg.id;
                for (var r =0; r < inJson.recipients.length; r++) {
                    var recipient = inJson.recipients[r];
                    var recip = models.MsgRecipient.build({
                        msgId: id,
                        clinicId: recipient.clinicId
                    });
                    promises.push(recip.save())
                }
                models.sequelize.Promise.all(promises).then(function(children){
                    outJson.recipients = children;
                    res.send(outJson);
                }).catch(function(recipErr){
                    res.status(500).send(recipErr);
                });
            }).catch(function(msgErr){
                res.status(500).send(msgErr);
            });
        }

    });

    app.put('/api/message/:id', function(req, res) {
        if (req.params && req.params['id']){
            var inJson = req.body;
            var outJson = {};
            if (!inJson.recipients || !Array.isArray(inJson.recipients)) {
                res.status("400").send("At least one recipient is required")
            }
            else {
                models.Message.findOne({
                    where: { id: req.params['id'] }
                }).then(function(existingMessage){
                    existingMessage.message = inJson.message;
                    existingMessage.createdByUserId = inJson.createdByUserId;
                    existingMessage.keepAliveStart = inJson.keepAliveStart;
                    existingMessage.keepAliveEnd = inJson.keepAliveEnd;
                    existingMessage.save().then(function(newMessage){
                        outJson = newMessage;
                        models.MsgRecipient.destroy({
                            where: { msgId: newMessage.id }
                        }).then(function(count){
                            var promises = [];
                            for (var r =0; r < inJson.recipients.length; r++) {
                                var recipient = inJson.recipients[r];
                                var recip = models.MsgRecipient.build({
                                    msgId: id,
                                    clinicId: recipient.clinicId
                                });
                                promises.push(recip.save())
                            }
                            models.sequelize.Promise.all(promises).then(function(children){
                                outJson.recipients = children;
                                res.send(outJson);
                            }).catch(function(recipErr){
                                res.status(500).send(recipErr);
                            });
                        }).catch(function(delErr){
                            res.status(500).send(delErr);
                        });
                    }).catch(function(saveErr){
                        res.status(500).send(saveErr);
                    });
                }).catch(function(msgErr){
                    res.status(500).send(msgErr);
                });
            }
        } else {
            res.status(400).send("Invalid id");
        }
    });

    //Special API for add or update functionality
    app.post('/api/v2/saveMessage', function(req, res) {
        console.log("request: " + req.session);
        upsertMessage(req.body, req)
            .then(function(result) {
                if (!result) result = req.body;
                res.send(result);
            });
    });

    app.post('/api/v2/deleteMessage', function(req, res) {
        models.Message.destroy({
            where: { id: req.body.id }
        }).then(function (){
            transactionLogger.info("Message deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
            res.send(req.body);
        });
    });



    /**
     *  Update or Insert a record. Searches for a record, if found, updates it. If not found, creates it then updates it.
     *  Returns 1 if updated, 0 if created, -1 for error (but not really)
     */
    function upsertMessage(message, req) {
        if (message.hasOwnProperty("recipients")) delete message.recipients;
        return models.Message.findOrCreate({where: {id: message.id}, defaults: {message: message.message, createdByUserId: message.createdByUserId}, raw: true })
            .spread(function(itemRecord, created) {
                message.id = itemRecord.id;
                models.Message.update(message, {where: {id: message.id}})
                    .then(function (count) {
                        if (count > 0 && message.Recipients) {
                            var ids = [];
                            for (var r=0; r < message.Recipients.length;r++){
                                ids.push(message.Recipients[r].id);
                            }
                            models.MsgRecipient.destroy({where: {messageId: message.id}}).then(function (num) {
                                var cats = [];
                                for (var i = 0; i < ids.length; i++) {
                                    var cat = models.MsgRecipient.build({
                                        messageId: message.id,
                                        clinicId: ids[i]
                                    });
                                    cats.push(cat);
                                }
                                if (cats.length > 0) {
                                    var promises = [];

                                    cats.forEach(function (cat) {
                                        promises.push(cat.save());
                                    });

                                    models.sequelize.Promise.all(promises).then(function (vals) {
                                        transactionLogger.info("Message " + ((created) ? "created" : "updated") + " by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(message));
                                        return cats.length;
                                    }).catch(function (saveErr) {
                                        logger.error(saveErr);
                                        return -1;
                                    });
                                }
                            }).catch(function (err) {
                                logger.error(err);
                                return -1;
                            });
                        }
                    }).catch(function (err) {
                    logger.error(err);
                    return -1;
                });
            }).catch(function(err) {
                logger.error(err);
                return -1;
            });
    }

    /**
     * Processes NDC scan JSON as a list
     *
     * @param dataObjects
     * @returns {*}
     */
    function bulkCreateNdcData(dataObjects) {

        var promises = [];

        if (dataObjects != null) {
            dataObjects.forEach(function (ndcData) {
                //logger.debug("saving ndc record: " + JSON.stringify(ndcData));
                ndcData.id = null;
                promises.push(models.NdcData.findOrCreate({
                        where: {
                            deviceId: ndcData.deviceId,
                            timestamp: ndcData.timestamp
                        }, raw: true
                    }).spread(function (fd, created) {
                                ndcData.id = fd.id;
                                models.NdcData.update(ndcData, {where: {id: ndcData.id}});
                                return {id : ndcData.id};
                            },
                            function (err) {
                                logger.error(err);
                            }
                        )
                );
            });
        }
        return models.sequelize.Promise.all(promises);
    }

    app.post('/api/saveNdcData', function(req, res) {
        bulkCreateNdcData(req.body)
            .done(function(data) {
                res.send(data);
            });
    });



    function buildWhereArgs(req, objName, objRawAttributes) {
        var whereArgs = [];
        if (req.body.where != null) {
            console.log(JSON.stringify(req.body.where));
            for (var key in req.body.where) {
                if (objRawAttributes.hasOwnProperty(key)) {
                    var newCondition = String(req.body.where[key]);
                    if (newCondition === "[object Object]") {
                        if (key === "dateModified") {
                            //we have gt/lt range
                            if (req.body.where[key]['$gt']) {
                                var newWhere = {};
                                newWhere[objRawAttributes[key].field] = {$gt: moment(req.body.where[key]['$gt'])};
                                whereArgs.push(newWhere);
                            }
                            if (req.body.where[key]['$lt']) {
                                var newWhere = {};
                                newWhere[objRawAttributes[key].field] = {$lt: moment(req.body.where[key]['$lt'])};
                                whereArgs.push(newWhere);
                            }
                            continue;
                        } else {
                            newCondition = String(req.body.where[key]['$like']);
                        }
                    }
                    var dataType = String(objRawAttributes[key].type);
                    if (newCondition.indexOf("*") < 0) {
                        newCondition = "%" + newCondition + "%";
                    } else {
                        newCondition = newCondition.replace(/\*/g, "%");
                    }
                    if (dataType.indexOf("VARCHAR") < 0) {
                        //cast numerics as strings so we can use the "LIKE" operator
                        if (dataType.indexOf("TIME") > 0) {
                            //since time is hidden in the UI, we should strip time from the date fields
                            whereArgs.push(models.Sequelize.where(models.Sequelize.cast(models.Sequelize.cast(models.Sequelize.col(objName + "." + objRawAttributes[key].field), 'DATE'), 'varchar'), 'LIKE', newCondition));
                        } else {
                            whereArgs.push(models.Sequelize.where(models.Sequelize.cast(models.Sequelize.col(objName + "." + objRawAttributes[key].field), 'varchar'), 'LIKE', newCondition));
                        }
                        delete req.body.where[key];
                    } else {
                        //already a String... just use the LIKE operator
                        var newWhere = {};

                        newWhere[objRawAttributes[key].field] = {$like: newCondition};
                        whereArgs.push(newWhere);
                    }
                }
            };
        }
        return whereArgs;
    }

    function buildMasterItemWhereArgs(req) {
        return buildWhereArgs(req, 'MasterItem', models.MasterItem.rawAttributes);
    }
    function buildClinicProfileWhereArgs(req) {
        return buildWhereArgs(req, 'ClinicProfile', models.ClinicProfile.rawAttributes);
    }

};


